/* Cross-references to other Sphinx objects */
div.admonition tt.xref, div.admonition code.xref, div.admonition a tt {
    background-color: #EEE;
    border-bottom: 1px dotted #888;
}

/* Inline code (``foo``) within docstring */
tt, code {
    background-color: #FFF;
    color: #222;
}

/* Table in automodule table of contents */
table.docutils {
    border-width: 0;
}

/* Table cells in automodule table of contents */
table.docutils td {
    border-width: 1px 0 1px 0;
    border-style: solid;
    border-color: #888;
    /* min-width: 300px; */
    padding: 5px;
}

div.figure {
    margin-right: 250px;
}

div#notebook p, div#notebook li {
    font-size: 17px;
}

div#notebook code, div#notebook pre {
    font-size: 15.3px;
}
